from pm4pymdl.algo.mvp.utils import succint_stream_to_exploded_stream
import pandas as pd


def apply(df):
    stream = df.to_dict('r')

    exploded_stream = succint_stream_to_exploded_stream.apply(stream)

    df = pd.DataFrame(exploded_stream)
    df.type = "exploded"

    return df



import pm4py
from pm4pymdl.algo.mvp.utils import succint_mdl_to_exploded_mdl
from pm4pymdl.algo.mvp.projection import factory
from pm4py.algo.discovery.inductive import factory as inductive_miner
from pm4py.algo.conformance.tokenreplay import factory as tr_factory
from pm4py.visualization.petrinet.util import performance_map
from pm4py.algo.filtering.log.variants import variants_filter as variants_module

PARAM_ACTIVITY_KEY = pm4py.util.constants.PARAMETER_CONSTANT_ACTIVITY_KEY

def apply(df, parameters=None):
    if df.type == "succint":
        df = succint_mdl_to_exploded_mdl(df)
        df.type = "exploded"

    persps = [x for x in df.columns if not x.startswith("event_")]

    ret = {}
    ret["nets"] = {}
    ret["act_count"] = {}
    ret["replay"] = {}
    ret["group_size_hist"] = {}
    ret["act_count_replay"] = {}
    ret["group_size_hist_replay"] = {}

    for persp in persps:
        print(persp,"getting log")
        log = factory.apply(df, persp, parameters=parameters)
        print(persp,"got log")

        net, im, fm = inductive_miner.apply(log)
        print(persp,"got model")

        activ_count = factory.apply(df, persp, variant="activity_occurrence", parameters=parameters)
        print(persp,"got activ_count")

        variants_idx = variants_module.get_variants_from_log_trace_idx(log)
        variants = variants_module.convert_variants_trace_idx_to_trace_obj(log, variants_idx)
        parameters_tr = {PARAM_ACTIVITY_KEY: "concept:name", "variants": variants}

        print(persp,"got variants")


        aligned_traces, place_fitness_per_trace, transition_fitness_per_trace, notexisting_activities_in_model = tr_factory.apply(log, net, im, fm, parameters={"enable_place_fitness": True, "disable_variants": True})

        print(persp,"done tbr")

        element_statistics = performance_map.single_element_statistics(log, net, im,
                                                                       aligned_traces, variants_idx)

        print(persp,"done element_statistics")

        aggregated_statistics = performance_map.aggregate_statistics(element_statistics)

        print(persp,"done aggregated_statistics")

        group_size_hist = factory.apply(df, persp, variant="group_size_hist", parameters=parameters)

        print(persp,"done group_size_hist")

        occurrences = {}
        for trans in transition_fitness_per_trace:
            occurrences[trans.label] = []
            for trace in transition_fitness_per_trace[trans]["fit_traces"]:
                if not trace in transition_fitness_per_trace[trans]["underfed_traces"]:
                    case_id = trace.attributes["concept:name"]
                    for event in trace:
                        if event["concept:name"] == trans.label:
                            occurrences[trans.label].append([case_id, event["event_id"]])
            #print(transition_fitness_per_trace[trans])


        len_different_ids = {}
        for act in occurrences:
            len_different_ids[act] = len(set(x[1] for x in occurrences[act]))

        eid_acti_count = {}
        for act in occurrences:
            eid_acti_count[act] = {}
            for x in occurrences[act]:
                if not x[0] in eid_acti_count:
                    eid_acti_count[act][x[0]] = 0
                eid_acti_count[act][x[0]] = eid_acti_count[act][x[0]] + 1
            eid_acti_count[act] = sorted(list(eid_acti_count[act].values()))

        ret["nets"][persp] = [net, im, fm]
        ret["act_count"][persp] = activ_count
        ret["replay"][persp] = aggregated_statistics
        ret["group_size_hist"][persp] = group_size_hist
        ret["act_count_replay"][persp] = len_different_ids
        ret["group_size_hist_replay"][persp] = eid_acti_count

        #print(ret["act_count"])
        #print(ret["act_count_replay"])


    return ret



from pm4py.objects.log.exporter.parquet import factory as parquet_exporter
from pm4pymdl.algo.mvp.utils import exploded_mdl_to_succint_mdl, succint_mdl_to_exploded_mdl
from copy import deepcopy

def apply(df, file_path, parameters=None):
    if parameters is None:
        parameters = {}

    if file_path.endswith(".csv") or file_path.endswith(".mdl"):
        conversion_needed = True
        try:
            if df.type == "succint":
                conversion_needed = False
        except:
            pass

        if conversion_needed:
            df = exploded_mdl_to_succint_mdl.apply(df)

        df.to_csv(file_path, index=False, sep=',', quotechar='\"')
    else:
        conversion_needed = False
        try:
            if df.type == "succint":
                conversion_needed = True
        except:
            pass

        if conversion_needed:
            df = succint_mdl_to_exploded_mdl.apply(df)
        new_parameters = deepcopy(parameters)
        new_parameters["compression"] = "gzip"
        parquet_exporter.export_df(df, file_path, parameters=new_parameters)



from datetime import datetime
from copy import deepcopy
from random import randrange, choice
import pandas as pd
from collections import Counter
import random

class shared:
    timestamp = 10000000
    # create N orders and deliveries, with 4*N packages
    N = 100
    event_count = 0
    # deliveries hesite
    esito = ["deliver", "undeliver"]


def generate_event(activity, related_classes):
    shared.event_count = shared.event_count + 1

    shared.timestamp = shared.timestamp + 1

    this_timestamp = datetime.fromtimestamp(shared.timestamp)

    base_event = {"event_id": "event_" + str(shared.event_count), "event_activity": activity,
                  "event_timestamp": this_timestamp}

    ret = []
    for cl in related_classes.keys():
        for el in related_classes[cl]:
            this_event = deepcopy(base_event)
            if cl == "order":
                this_event[cl] = "order_" + str(el)
            if cl == "package":
                this_event[cl] = "package_" + str(el)
            if cl == "delivery":
                this_event[cl] = "delivery_" + str(el)
            if cl == "back":
                this_event[cl] = "back_" + str(el)

            ret.append(this_event)

    return ret


def generate_log():
    list_events = []

    for d in range(shared.N*2):

        income = randrange(100, 1000)
        costs = randrange(0, income)
        profit = income - costs

        event = generate_event("create", {"order": [d]})

        for i in range(len(event)):
            event[i]["event_income"] = income
            event[i]["event_costs"] = costs
            event[i]["event_profit"] = profit

        list_events = list_events + event

    i = 0
    while i < shared.N*2:

        list_events = list_events + generate_event("split", {"package": [8*i, 8*i +1, 8*i +2, 8*i +3], "order": [i]})
        list_events = list_events + generate_event("split", {"package": [8*i + 4, 8*i +5, 8*i +6, 8*i +7], "order": [i+1]})

        i = i + 2

    i = 0
    while i < shared.N*2:
        list_events = list_events + generate_event("load", {"package": [8*i, 8*i +1, 8*i +4, 8*i +5], "delivery": [i]})
        list_events = list_events + generate_event("load", {"package": [8*i +2, 8*i +3, 8*i +6, 8*i +7], "delivery": [i+1]})

        i = i + 2

    i = 0
    while i < shared.N*2:
        j = 0
        while True:
            esito = choice(shared.esito)
            j = j + 1

            list_events = list_events + generate_event(esito, {"package": [8*i, 8*i +1, 8*i +4, 8*i +5], "delivery": [i]})
            list_events = list_events + generate_event(esito, {"package": [8*i +2, 8*i +3, 8*i +6, 8*i +7], "delivery": [i+1]})

            list_events = list_events + generate_event("next",
                                                       {
                                                        "delivery": [i]})
            list_events = list_events + generate_event("next",
                                                       {
                                                        "delivery": [i + 1]})

            if not esito == "deliver":
                r = random.random()
                if r < 0.5:
                    list_events = list_events + generate_event("retry",
                                                               {"package": [8 * i, 8 * i + 1, 8 * i + 4, 8 * i + 5],
                                                                "delivery": [i]})
                    list_events = list_events + generate_event("retry",
                                                               {"package": [8 * i + 2, 8 * i + 3, 8 * i + 6, 8 * i + 7],
                                                                "delivery": [i + 1]})
                    continue
                else:
                    list_events = list_events + generate_event("return", {"package": [8 * i, 8 * i + 1, 8 * i + 4, 8 * i + 5, 8*i +2, 8*i +3, 8*i +6, 8*i +7]})
                    break

        i = i + 2

    for d in range(shared.N*2):

        list_events = list_events + generate_event("notify", {"order": [d]})

    i = 0
    while i < shared.N*2:
        additional_fee = randrange(0, 200)
        event = generate_event("bill", {"package": [8*i, 8*i +1, 8*i +2, 8*i +3], "order": [i]})
        for j in range(len(event)):
            event[j]["event_add_fee"] = additional_fee
        list_events = list_events + event

        event = generate_event("bill", {"package": [8*i + 4, 8*i +5, 8*i +6, 8*i +7], "order": [i+1]})
        for j in range(len(event)):
            event[j]["event_add_fee"] = additional_fee
        list_events = list_events + event

        i = i + 2

    i = 0
    while i < shared.N*2:
        list_events = list_events + generate_event("finish", {"package": [8*i, 8*i +1, 8*i +4, 8*i +5], "delivery": [i]})
        list_events = list_events + generate_event("finish", {"package": [8*i +2, 8*i +3, 8*i +6, 8*i +7], "delivery": [i+1]})

        i = i + 2

    df = pd.DataFrame(list_events)
    df.type = "exploded"

    return df


if __name__ == "__main__":
    df = generate_log()

    print(df)



def apply(df, remove_common=False, include_activity_timest_in_key=False):
    """
    Dataframe to grouped stream

    Parameters
    -------------
    df
        Dataframe

    Returns
    -------------
    grouped_stream
        Grouped stream of events
    """
    stream = df.to_dict('r')
    grouped_stream = {}

    i = 0
    while i < len(stream):
        keys = list(stream[i].keys())
        for key in keys:
            if str(stream[i][key]) == "nan" or str(stream[i][key]) == "None":
                del stream[i][key]
        if "event_id" in stream[i] and "event_activity" in stream[i] and "event_timestamp" in stream[i]:
            event_id = stream[i]["event_id"]
            event_activity = stream[i]["event_activity"]
            event_timestamp = stream[i]["event_timestamp"]
            if include_activity_timest_in_key:
                if (event_id, event_activity, event_timestamp) not in grouped_stream:
                    grouped_stream[(event_id, event_activity, event_timestamp)] = []
            else:
                if event_id not in grouped_stream:
                    grouped_stream[event_id] = []
            if remove_common:
                del stream[i]["event_id"]
                del stream[i]["event_activity"]
                if "event_timestamp" in stream[i]:
                    del stream[i]["event_timestamp"]
            if include_activity_timest_in_key:
                grouped_stream[(event_id, event_activity, event_timestamp)].append(stream[i])
            else:
                grouped_stream[event_id].append(stream[i])
        i = i + 1

    return grouped_stream



from pm4pymdl import algo, objects, visualization, order_log_generation

__version__ = '0.0.6'
__doc__ = "Process Mining for Python - Multi-Dimensional Event Logs"
__author__ = 'PADS'
__author_email__ = 'a.berti@pads.rwth-aachen.de'
__maintainer__ = 'PADS'
__maintainer_email__ = "a.berti@pads.rwth-aachen.de"



from pm4pymdl.algo.mvp.gen_framework.util import filter_on_map
import pandas as pd


def apply(df, model, rel_ev, rel_act, parameters=None):
    if parameters is None:
        parameters = {}

    new_df = filter_on_map.apply(df, model.map)

    df_list = []
    for persp in rel_ev:
        proj_df = df[["event_id", "event_activity", persp]].dropna()
        proj_df = proj_df.groupby(["event_id", persp]).first()
        df_list.append(proj_df)

    concat_df = pd.concat(df_list)

    return dict(concat_df["event_activity"].value_counts())



from pm4pymdl.algo.mvp.gen_framework.rel_events import rel_dfg
from pm4pymdl.algo.mvp.gen_framework.rel_events import being_produced
from pm4pymdl.algo.mvp.gen_framework.rel_events import existence
from pm4pymdl.algo.mvp.gen_framework.rel_events import link
from pm4pymdl.algo.mvp.gen_framework.rel_events import eventually_follows

REL_DFG = "rel_dfg"
BEING_PRODUCED = "being_produced"
EXISTENCE = "existence"
LINK = "link"
EVENTUALLY_FOLLOWS = "eventually_follows"

VERSIONS = {REL_DFG: rel_dfg.apply, BEING_PRODUCED: being_produced.apply, EXISTENCE: existence.apply, LINK: link.apply,
            EVENTUALLY_FOLLOWS: eventually_follows.apply}


def apply(df, model, variant=REL_DFG, parameters=None):
    if parameters is None:
        parameters = {}

    return VERSIONS[variant](df, model, parameters=parameters)



import uuid
import tempfile
from graphviz import Digraph
from pm4py.objects.petri.petrinet import PetriNet


COLORS = ["#05B202", "#A13CCD", "#39F6C0", "#BA0D39", "#E90638", "#07B423", "#306A8A", "#678225", "#2742FE", "#4C9A75",
          "#4C36E9", "#7DB022", "#EDAC54", "#EAC439", "#EAC439", "#1A9C45", "#8A51C4", "#496A63", "#FB9543", "#2B49DD",
          "#13ADA5", "#2DD8C1", "#2E53D7", "#EF9B77", "#06924F", "#AC2C4D", "#82193F", "#0140D3"]


def apply(obj, parameters=None):
    if parameters is None:
        parameters = {}

    image_format = "png"
    if "format" in parameters:
        image_format = parameters["format"]

    filename = tempfile.NamedTemporaryFile(suffix='.gv').name
    g = Digraph("", filename=filename, engine='dot', graph_attr={'bgcolor': 'transparent'})

    nets = obj["nets"]
    replay = obj["replay"]
    act_count = obj["act_count_replay"]
    group_size = obj["group_size_hist_replay"]

    all_objs = {}
    trans_names = {}


    for index, persp in enumerate(nets):
        orig_act_count = {}
        orig_arc_count = {}

        color = COLORS[index % len(COLORS)]
        net, im, fm = nets[persp]
        rr = replay[persp]
        ac = act_count[persp]
        gs = group_size[persp]

        for pl in net.places:
            this_uuid = str(uuid.uuid4())
            if pl in im:
                g.node(this_uuid, "", shape="circle", fillcolor=color, style="filled")
            elif pl in fm:
                g.node(this_uuid, "", shape="circle", fillcolor=color, style="filled")
            else:
                g.node(this_uuid, "", shape="circle", fillcolor=color, style="filled")
            all_objs[pl] = this_uuid

        for tr in net.transitions:
            this_uuid = str(uuid.uuid4())
            if tr.label is None:
                g.node(this_uuid, "", shape="box", fillcolor="#000000", style="filled")
                all_objs[tr] = this_uuid
            elif tr.label not in trans_names:
                count = rr[tr]["label"].split("(")[-1].split("]")[0]
                orig_act_count[tr] = count
                this_act_count = "%.2f" % (float(ac[tr.label]))
                g.node(this_uuid, "'" + tr.label+" <<"+this_act_count+">>'", shape="box")
                trans_names[tr.label] = this_uuid
                all_objs[tr] = this_uuid
            else:
                all_objs[tr] = trans_names[tr.label]

        for arc in net.arcs:
            this_uuid = str(uuid.uuid4())

            source_node = arc.source
            target_node = arc.target

            if arc in rr:

                arc_count = float(rr[arc]['label'])

                if type(source_node) is PetriNet.Place:
                    if target_node.label is not None and ac[target_node.label] > 0:
                        ratio = "'<<%.2f>>'" % (arc_count / ac[target_node.label])
                    else:
                        ratio = "'<<1.00>>'"
                else:
                    if source_node.label is not None and ac[source_node.label] > 0:
                        ratio = "'<<%.2f>>'" % (arc_count / ac[source_node.label])
                    else:
                        ratio = "'<<1.00>>'"

                g.edge(all_objs[source_node], all_objs[target_node], label=ratio, penwidth=str(rr[arc]['penwidth']), color=color)

                all_objs[arc] = this_uuid
            else:
                g.edge(all_objs[source_node], all_objs[target_node], label="", color=color)

    g.attr(overlap='false')
    g.attr(fontsize='11')

    g.format = image_format

    return g



from pm4pymdl.algo.mvp.utils import exploded_mdl_to_stream
import pandas as pd


def apply(df):
    stream = exploded_mdl_to_stream.apply(df)

    df = pd.DataFrame(stream)
    df.type = "succint"

    return df



from pm4pymdl.algo.mvp.discovery.versions import classic

CLASSIC = "classic"

VERSIONS = {CLASSIC: classic.apply}


def apply(dataframe, parameters=None, variant=CLASSIC):
    """
    Discover a StarStar model from an ad-hoc built dataframe

    Parameters
    -------------
    df
        Dataframe
    parameters
        Possible parameters of the algorithm
    variant
        Variant of the algorithm, possible values: classic

    Returns
    -------------
    perspectives_heu
        Dictionary of perspectives associated to Heuristics Net
    """
    return VERSIONS[variant](dataframe, parameters=parameters)



from pm4py.objects.log.importer.parquet import factory as parquet_importer
import pandas as pd

def apply(file_path, parameters=None):
    if parameters is None:
        parameters = {}

    if file_path.endswith(".csv") or file_path.endswith(".mdl"):
        df = pd.read_csv(file_path, sep=',', quotechar='\"')
        df.type = "succint"
        return df
    elif file_path.endswith(".parquet"):
        df = parquet_importer.apply(file_path)
        df.type = "exploded"
        return df



from copy import copy
from datetime import datetime
from dateutil import parser

import pandas as pd

import random


def apply(file_path, parameters=None):
    """
    Apply the importing of a XOC file

    Parameters
    ------------
    file_path
        Path to the XOC file
    parameters
        Import parameters

    Returns
    ------------
    dataframe
        Dataframe
    """
    if parameters is None:
        parameters = {}

    import_timestamp = parameters["import_timestamp"] if "import_timestamp" in parameters else True
    sample_probability = parameters["sample_probability"] if "sample_probability" in parameters else None
    classes_to_reference = set(parameters["classes_to_reference"]) if "classes_to_reference" in parameters else None
    if classes_to_reference is None:
        classes_to_reference = set()

    F = open(file_path, "r")
    content = F.read()
    F.close()
    events = content.split("<event>")
    stream = []
    stream_strings = []
    i = 1
    considered_events = 0
    considered_objects = set()
    considered_activities = set()
    considered_classes = set()
    while i < len(events) - 1:
        if sample_probability is not None:
            r = random.random()
            if r > sample_probability:
                i = i + 1
                continue
        considered_events = considered_events + 1
        event_id = events[i].split("\"id\" value=\"")[1].split("\"")[0]
        event_activity = events[i].split("\"activity\" value=\"")[1].split("\"")[0]
        considered_activities.add(event_activity)
        event_timestamp0 = events[i].split("\"timestamp\" value=\"")[1].split("\"")[0].replace(" CET", "")
        event_timestamp = None
        if import_timestamp:
            try:
                event_timestamp = parser.parse(event_timestamp0)
            except:
                pass
        if event_timestamp is not None:
            event_dictio = {"event_id": event_id, "event_activity": event_activity, "event_timestamp": event_timestamp}
        else:
            event_dictio = {"event_id": event_id, "event_activity": event_activity}
        references = events[i].split("<references>")[1].split("</references>")[0].split("<object>")
        referenced_classes = set()
        j = 1
        while j < len(references):
            object_class = references[j].split("\"class\" value=\"")[1].split("\"")[0]
            referenced_classes.add(object_class)
            j = j + 1
        if referenced_classes.issuperset(classes_to_reference):
            j = 1
            while j < len(references):
                this_event_dictio = copy(event_dictio)
                object_id = references[j].split("\"id\" value=\"")[1].split("\"")[0]
                considered_objects.add(object_id)
                object_class = references[j].split("\"class\" value=\"")[1].split("\"")[0]
                considered_classes.add(object_class)
                this_event_dictio[object_class] = object_id
                this_event_dictio_stri = str(this_event_dictio)
                if this_event_dictio_stri not in stream_strings:
                    stream.append(this_event_dictio)
                    stream_strings.append(this_event_dictio_stri)
                j = j + 1
        i = i + 1
    dataframe = pd.DataFrame.from_dict(stream)
    if import_timestamp:
        dataframe = dataframe.sort_values(["event_timestamp", "event_id"])

    dataframe.type = "exploded"

    print("events: ",considered_events,"objects: ",len(considered_objects),"activities: ",len(considered_activities),"classes: ",len(considered_classes))
    return dataframe



from pm4pymdl.algo.mvp.gen_framework.node_freq import type1, type21, type22, type23, type31, type32, type33

TYPE1 = "type1"
TYPE21 = "type21"
TYPE22 = "type22"
TYPE23 = "type23"
TYPE31 = "type31"
TYPE32 = "type32"
TYPE33 = "type33"

VERSIONS = {TYPE1: type1.apply, TYPE21: type21.apply, TYPE22: type22.apply, TYPE23: type23.apply,
            TYPE31: type31.apply, TYPE32: type32.apply, TYPE33: type33.apply}


def apply(df, model, rel_ev, rel_act, variant=TYPE1, parameters=None):
    return VERSIONS[variant](df, model, rel_ev, rel_act, parameters=parameters)



from pm4pymdl.objects.xoc.exporter.versions import classic

CLASSIC = "classic"

VERSIONS = {CLASSIC: classic.apply}


def apply(df, dest_path, parameters=None, variant=CLASSIC):
    """
    Export into a XOC format

    Parameters
    ------------
    df
        Dataframe
    dest_path
        Destination path
    parameters
        Parameters of the algorithm
    variant
        Variant of the algorithm, possible values: classic
    """
    return VERSIONS[variant](df, dest_path, parameters=parameters)



import pandas as pd


def apply(df, model, parameters=None):
    if parameters is None:
        parameters = {}

    ret = {}
    columns = [x for x in df.columns if not x.startswith("event_")]
    basic_columns = ["event_id", "event_activity"]
    for col in columns:
        proj_df_0 = df[basic_columns + [col]].dropna()
        proj_df_0 = proj_df_0.groupby(col).first().reset_index()
        #proj_df = proj_df_0.groupby("event_id").first().reset_index()
        #proj_df_2 = proj_df_0.groupby("event_id").first().reset_index()
        proj_df = proj_df_0.copy()
        proj_df_2 = proj_df_0.copy()
        proj_df_2.columns = [x+"_2" for x in proj_df_2.columns]
        concat_df = pd.concat([proj_df, proj_df_2], axis=1)
        concat_df["event_activity_merge"] = concat_df["event_activity"] + "@@" + concat_df["event_activity_2"]
        #concat_df["event_id_2"] = concat_df["event_id"]
        ret[col] = concat_df

    return ret



from pm4pymdl.algo.mvp.gen_framework.util import filter_on_map

def apply(df, model, rel_ev, rel_act, parameters=None):
    if parameters is None:
        parameters = {}

    new_df = filter_on_map.apply(df, model.map)

    return dict(new_df["event_activity"].value_counts())



from pm4pymdl.algo.mvp.utils import exploded_mdl_to_grouped_stream_old, exploded_mdl_to_stream, \
    exploded_mdl_to_succint_mdl, generator, succint_mdl_to_exploded_mdl, succint_stream_to_exploded_stream




def apply(df, model, rel_ev, rel_act, parameters=None):
    if parameters is None:
        parameters = {}

    ret = {}
    list_persp = [x for x in df.columns if not x.startswith("event_")]

    for p in list_persp:
        proj_df = df[["event_id", "event_activity", p]].dropna()

        ret[p] = dict(proj_df.groupby(["event_id", p]).first()["event_activity"].value_counts())

    return ret



import math
import tempfile

import pydotplus

from pm4py.visualization.common.utils import human_readable_stat


def get_corr_hex(num):
    """
    Gets correspondence between a number
    and an hexadecimal string

    Parameters
    -------------
    num
        Number

    Returns
    -------------
    hex_string
        Hexadecimal string
    """
    if num < 10:
        return str(int(num))
    elif num < 11:
        return "A"
    elif num < 12:
        return "B"
    elif num < 13:
        return "C"
    elif num < 14:
        return "D"
    elif num < 15:
        return "E"
    elif num < 16:
        return "F"


def transform_to_hex(graycolor):
    """
    Transform color to hexadecimal representation

    Parameters
    -------------
    graycolor
        Gray color (int from 0 to 255)

    Returns
    -------------
    hex_string
        Hexadecimal color
    """
    left0 = graycolor / 16
    right0 = graycolor % 16

    left00 = get_corr_hex(left0)
    right00 = get_corr_hex(right0)

    return "#" + left00 + right00 + left00 + right00 + left00 + right00


def transform_to_hex_2(color):
    """
    Transform color to hexadecimal representation

    Parameters
    -------------
    color
        Gray color (int from 0 to 255)

    Returns
    -------------
    hex_string
        Hexadecimal color
    """
    color = 255 - color
    color2 = 255 - color

    left0 = color / 16
    right0 = color % 16

    left1 = color2 / 16
    right1 = color2 % 16

    left0 = get_corr_hex(left0)
    right0 = get_corr_hex(right0)
    left1 = get_corr_hex(left1)
    right1 = get_corr_hex(right1)

    return "#" + left0 + right0 + left1 + right1 + left1 + right1

def apply_vis_heu(heu_net, parameters=None):
    """
    Gets a representation of an Heuristics Net

    Parameters
    -------------
    heu_net
        Heuristics net
    parameters
        Possible parameters of the algorithm, including: format

    Returns
    ------------
    gviz
        Representation of the Heuristics Net
    """
    if parameters is None:
        parameters = {}

    image_format = parameters["format"] if "format" in parameters else "png"

    graph = pydotplus.Dot(strict=True)
    graph.obj_dict['attributes']['bgcolor'] = 'transparent'

    corr_nodes = {}
    corr_nodes_names = {}
    is_frequency = False

    for node_name in heu_net.nodes:
        node = heu_net.nodes[node_name]
        node_occ = node.node_occ
        graycolor = transform_to_hex_2(max(255 - math.log(node_occ) * 9, 0))
        if node.node_type == "frequency":
            is_frequency = True
            n = pydotplus.Node(name=node_name, shape="box", style="filled",
                               label=node_name + " (" + str(node_occ) + ")", fillcolor=graycolor)
        else:
            n = pydotplus.Node(name=node_name, shape="box", style="filled",
                               label=node_name, fillcolor=graycolor)
        corr_nodes[node] = n
        corr_nodes_names[node_name] = n
        graph.add_node(n)

    # gets max arc value
    max_arc_value = -1
    for node_name in heu_net.nodes:
        node = heu_net.nodes[node_name]
        for other_node in node.output_connections:
            if other_node in corr_nodes:
                for edge in node.output_connections[other_node]:
                    max_arc_value = max(max_arc_value, edge.repr_value)

    for node_name in heu_net.nodes:
        node = heu_net.nodes[node_name]
        for other_node in node.output_connections:
            if other_node in corr_nodes:
                for edge in node.output_connections[other_node]:
                    this_pen_width = 1.0 + math.log(1 + edge.repr_value) / 11.0
                    repr_value = str(edge.repr_value)
                    if edge.net_name:
                        if node.node_type == "frequency":
                            e = pydotplus.Edge(src=corr_nodes[node], dst=corr_nodes[other_node],
                                               label=edge.net_name + " (" + repr_value + ")",
                                               color=edge.repr_color,
                                               fontcolor=edge.repr_color, penwidth=this_pen_width)
                        else:
                            e = pydotplus.Edge(src=corr_nodes[node], dst=corr_nodes[other_node],
                                               label=edge.net_name + " (" + human_readable_stat(repr_value) + ")",
                                               color=edge.repr_color,
                                               fontcolor=edge.repr_color, penwidth=this_pen_width)
                    else:
                        if node.node_type == "frequency":
                            e = pydotplus.Edge(src=corr_nodes[node], dst=corr_nodes[other_node], label=repr_value,
                                               color=edge.repr_color,
                                               fontcolor=edge.repr_color, penwidth=this_pen_width)
                        else:
                            e = pydotplus.Edge(src=corr_nodes[node], dst=corr_nodes[other_node],
                                               label=human_readable_stat(repr_value),
                                               color=edge.repr_color,
                                               fontcolor=edge.repr_color, penwidth=this_pen_width)

                    graph.add_edge(e)

    for index, sa_list in enumerate(heu_net.start_activities):
        effective_sa_list = [n for n in sa_list if n in corr_nodes_names]
        if effective_sa_list:
            start_i = pydotplus.Node(name="start_" + str(index), label="@@S", color=heu_net.default_edges_color[index],
                                     fontsize="8", fontcolor="#32CD32", fillcolor="#32CD32",
                                     style="filled")
            graph.add_node(start_i)
            for node_name in effective_sa_list:
                sa = corr_nodes_names[node_name]
                if type(heu_net.start_activities[index]) is dict:
                    if is_frequency:
                        occ = heu_net.start_activities[index][node_name]
                        this_pen_width = 1.0 + math.log(1 + occ) / 11.0
                        if heu_net.net_name[index]:
                            e = pydotplus.Edge(src=start_i, dst=sa, label=heu_net.net_name[index] + " (" + str(occ) + ")",
                                               color=heu_net.default_edges_color[index],
                                               fontcolor=heu_net.default_edges_color[index], penwidth=this_pen_width)
                        else:
                            e = pydotplus.Edge(src=start_i, dst=sa, label=str(occ),
                                               color=heu_net.default_edges_color[index],
                                               fontcolor=heu_net.default_edges_color[index], penwidth=this_pen_width)
                    else:
                        e = pydotplus.Edge(src=start_i, dst=sa, label=heu_net.net_name[index],
                                           color=heu_net.default_edges_color[index],
                                           fontcolor=heu_net.default_edges_color[index])
                else:
                    e = pydotplus.Edge(src=start_i, dst=sa, label=heu_net.net_name[index],
                                       color=heu_net.default_edges_color[index],
                                       fontcolor=heu_net.default_edges_color[index])
                graph.add_edge(e)

    for index, ea_list in enumerate(heu_net.end_activities):
        effective_ea_list = [n for n in ea_list if n in corr_nodes_names]
        if effective_ea_list:
            end_i = pydotplus.Node(name="end_" + str(index), label="@@E", color="#",
                                   fillcolor="#FFA500", fontcolor="#FFA500", fontsize="8",
                                   style="filled")
            graph.add_node(end_i)
            for node_name in effective_ea_list:
                ea = corr_nodes_names[node_name]
                if type(heu_net.end_activities[index]) is dict:
                    if is_frequency:
                        occ = heu_net.end_activities[index][node_name]
                        this_pen_width = 1.0 + math.log(1 + occ) / 11.0
                        if heu_net.net_name[index]:
                            e = pydotplus.Edge(src=ea, dst=end_i, label=heu_net.net_name[index] + " (" + str(occ) + ")",
                                               color=heu_net.default_edges_color[index],
                                               fontcolor=heu_net.default_edges_color[index], penwidth=this_pen_width)
                        else:
                            e = pydotplus.Edge(src=ea, dst=end_i, label=str(occ),
                                               color=heu_net.default_edges_color[index],
                                               fontcolor=heu_net.default_edges_color[index], penwidth=this_pen_width)
                    else:
                        e = pydotplus.Edge(src=ea, dst=end_i, label=heu_net.net_name[index],
                                           color=heu_net.default_edges_color[index],
                                           fontcolor=heu_net.default_edges_color[index])
                else:
                    e = pydotplus.Edge(src=ea, dst=end_i, label=heu_net.net_name[index],
                                       color=heu_net.default_edges_color[index],
                                       fontcolor=heu_net.default_edges_color[index])
                graph.add_edge(e)

    print("n. edges MVP",len(graph.get_edge_list()))
    file_name = tempfile.NamedTemporaryFile(suffix='.' + image_format)
    file_name.close()
    graph.write(file_name.name, format=image_format)


    return file_name


def apply(model, parameters=None):
    """
    Visualize StarStar models

    Parameters
    ------------
    model
        StarStar model
    parameters
        Possible parameters of the algorithm

    Returns
    -----------
    file_name
        File name where a representation of the StarStar model is stored
    """
    if parameters is None:
        parameters = {}

    overall_heu = None
    for persp_name in model:
        persp = model[persp_name]
        if overall_heu is None:
            overall_heu = persp
        else:
            overall_heu = overall_heu + persp
    if overall_heu is not None:
        return apply_vis_heu(overall_heu, parameters=parameters)
    return None


from graphviz import Digraph
import uuid
import tempfile
from pm4pymdl.visualization.mvp.gen_framework.versions import dfg_clean

COLORS = ["#05B202", "#A13CCD", "#39F6C0", "#BA0D39", "#E90638", "#07B423", "#306A8A", "#678225", "#2742FE", "#4C9A75",
          "#4C36E9", "#7DB022", "#EDAC54", "#EAC439", "#EAC439", "#1A9C45", "#8A51C4", "#496A63", "#FB9543", "#2B49DD",
          "#13ADA5", "#2DD8C1", "#2E53D7", "#EF9B77", "#06924F", "#AC2C4D", "#82193F", "#0140D3"]


def apply(model, parameters=None):
    if parameters is None:
        parameters = {}

    min_node_freq = parameters["min_node_freq"] if "min_node_freq" in parameters else 0
    min_edge_freq = parameters["min_edge_freq"] if "min_edge_freq" in parameters else 0
    dfg_cleaning_threshold = parameters["dfg_cleaning_threshold"] if "dfg_cleaning_threshold" in parameters else -1
    max_edge_ratio = parameters["max_edge_ratio"] if "max_edge_ratio" in parameters else 100000000

    image_format = "png"
    if "format" in parameters:
        image_format = parameters["format"]

    filename = tempfile.NamedTemporaryFile(suffix='.gv').name
    g = Digraph("", filename=filename, engine='dot', graph_attr={'bgcolor': 'transparent'})

    cluster_corr = {}
    cluster_acti_corr = {}

    count_nodes = 0

    perspectives = sorted(list(model.node_freq.keys()))
    for index, p in enumerate(perspectives):
        color = COLORS[index % len(COLORS)]

        with g.subgraph(name='cluster_' + p) as c:
            cluster_corr[p] = c
            cluster_acti_corr[p] = {}

            c.attr(style='filled')
            c.attr(color='lightgrey')
            c.node_attr.update(style='filled', color='white')
            c.attr(label='class: ' + p)

            for act in model.node_freq[p]:
                if model.node_freq[p][act] >= min_node_freq:
                    this_uuid = str(uuid.uuid4())
                    c.node(this_uuid, act + "(" + str(model.node_freq[p][act]) + ")", fillcolor=color, style="filled")
                    cluster_acti_corr[p][act] = this_uuid

                    count_nodes = count_nodes + 1

    count_edges = 0

    for index, persp in enumerate(perspectives):
        color = COLORS[index % len(COLORS)]
        edges = model.edge_freq[persp]

        if dfg_cleaning_threshold > -1:
            edges = dfg_clean.clean_dfg_based_on_noise_thresh(edges, model.node_freq, dfg_cleaning_threshold)

        for edge in edges:
            if edge.split("@@")[0] in cluster_acti_corr[persp] and edge.split("@@")[1] in cluster_acti_corr[persp]:
                if edges[edge] >= min_edge_freq:
                    act1 = edge.split("@@")[0]
                    act2 = edge.split("@@")[1]
                    act1_corr = cluster_acti_corr[persp][act1]
                    act2_corr = cluster_acti_corr[persp][act2]

                    if max(edges[edge] / model.node_freq[persp][act1],
                           edges[edge] / model.node_freq[persp][act2]) < max_edge_ratio:
                        g.edge(act1_corr, act2_corr, persp + " (" + str(edges[edge]) + ")", color=color,
                               fontcolor=color)
                        count_edges = count_edges + 1

    if "count_nodes" in parameters and "count_edges" in parameters:
        print("Type 3 nodes", count_nodes, "recall",count_nodes/parameters["count_nodes"])
        print("Type 3 edges", count_edges, "recall",count_edges/parameters["count_edges"])
    else:
        print("Type 3 nodes", count_nodes)
        print("Type 3 edges", count_edges)

    g.attr(overlap='false')
    g.attr(fontsize='11')

    g.format = image_format

    return g



from pm4pymdl.algo.mvp.gen_framework.models.basic_model_structure import BasicModelStructure


class MVPModel3(BasicModelStructure):
    def __init__(self, df, parameters=None):
        self.type = "model3"
        BasicModelStructure.__init__(self, df, parameters=parameters)



from pm4pymdl.algo.mvp.gen_framework import factory as gen_fram_factory
from pm4py.algo.discovery.inductive import factory as inductive_miner
from pm4py.algo.discovery.dfg.utils import dfg_utils


class ProclDerivationModel(object):
    def __init__(self, df, parameters=None):
        if parameters is None:
            parameters = {}

        self.type = "proclet_derivation_model"

        self.df = df
        self.model_dfg = gen_fram_factory.apply(df, model_type_variant="model3", rel_ev_variant="rel_dfg",
                                                node_freq_variant="type31", edge_freq_variant="type11")
        self.model_3_1 = gen_fram_factory.apply(df, model_type_variant="model3", rel_ev_variant="being_produced",
                                                node_freq_variant="type31", edge_freq_variant="type11")
        self.model_3_2 = gen_fram_factory.apply(df, model_type_variant="model3", rel_ev_variant="being_produced",
                                                node_freq_variant="type32", edge_freq_variant="type13")
        self.model_link = gen_fram_factory.apply(df, model_type_variant="model3", rel_ev_variant="link",
                                                 node_freq_variant="type32", edge_freq_variant="type11")

        self.model_inductive = {}
        self.start_activities = {}
        self.end_activities = {}

        for cl in self.model_dfg.edge_freq:
            this_dfg = self.model_dfg.edge_freq[cl]
            new_dfg = {(x.split("@@")[0], x.split("@@")[1]): this_dfg[x] for x in this_dfg}
            self.model_inductive[cl] = inductive_miner.apply_dfg(new_dfg)
            self.start_activities[cl] = dfg_utils.infer_start_activities(new_dfg)
            self.end_activities[cl] = dfg_utils.infer_end_activities(new_dfg)

        self.linked_perspectives = {}

        self.find_linked_perspectives()

    def find_linked_perspectives(self):
        self.act_count_1 = self.model_3_1.node_freq
        self.act_count_2 = self.model_3_2.node_freq

        self.acti_class_map = {}

        for cl in self.act_count_1:
            for act in self.act_count_1[cl]:
                if not act in self.acti_class_map:
                    self.acti_class_map[act] = set()
                self.acti_class_map[act].add(cl)

        self.loop_count_1 = {}
        self.loop_count_2 = {}
        self.acti_loop_map = {}
        self.production_map = {}

        for cl in self.model_3_1.edge_freq:
            self.loop_count_1[cl] = {}
            for edge in self.model_3_1.edge_freq[cl]:
                acti = edge.split("@@")[0]
                self.loop_count_1[cl][acti] = self.model_3_1.edge_freq[cl][edge]
                if not acti in self.acti_loop_map:
                    self.acti_loop_map[acti] = set()
                self.acti_loop_map[acti].add(cl)

        for cl in self.model_3_2.edge_freq:
            self.loop_count_2[cl] = {}
            for edge in self.model_3_2.edge_freq[cl]:
                self.loop_count_2[cl][edge.split("@@")[0]] = self.model_3_2.edge_freq[cl][edge]

        for acti in self.acti_loop_map:
            if len(self.acti_loop_map[acti]) == 1 and len(self.acti_class_map[acti]) > 1:
                current_class = list(self.acti_loop_map[acti])[0]
                remaining_classes = self.acti_class_map[acti] - self.acti_loop_map[acti]
                linked_acti_current = set(
                    (x,
                     self.model_3_2.node_freq[current_class][x.split("@@")[1]],
                     self.model_3_2.node_freq[current_class][x.split("@@")[0]]) for x in
                    self.model_link.edge_freq[current_class] if
                    x.split("@@")[0] == acti and self.model_link.edge_freq[current_class][x] >=
                    self.model_3_1.node_freq[current_class][x.split("@@")[1]])

                if linked_acti_current:
                    prod_count_1_source = self.loop_count_1[current_class][acti]
                    acti_count_1_source = self.act_count_1[current_class][acti]
                    prod_count_2_source = self.loop_count_2[current_class][acti]

                    if prod_count_1_source == acti_count_1_source:
                        if prod_count_2_source > prod_count_1_source:
                            source_card = "N"
                        else:
                            source_card = "1"
                    else:
                        if prod_count_2_source > prod_count_1_source:
                            source_card = "N"
                        else:
                            source_card = "1"

                    for cl in remaining_classes:
                        linked = None
                        linked = []

                        for x in linked_acti_current:
                            ta = x[0].split("@@")[1]
                            if ta in self.act_count_1[cl]:
                                if x[1] == x[2]:
                                    linked.append([x[0].split("@@")[1], "="])
                                else:
                                    linked.append([x[0].split("@@")[1], "<="])

                        if linked:
                            if current_class not in self.production_map:
                                self.production_map[current_class] = {}
                            if cl not in self.production_map[current_class]:
                                self.production_map[current_class][cl] = {}

                            acti_count_1_target = self.act_count_1[cl][acti]
                            acti_count_2_target = self.act_count_2[cl][acti]

                            if acti_count_2_target > acti_count_1_target:
                                target_card = "N"
                            else:
                                target_card = "1"

                            self.production_map[current_class][cl][acti] = {"linked": linked, "source_card": source_card,
                                                                            "target_card": target_card}
